@@ -18,11 +18,7 @@ specific language governing permissions and limitations
 under the License.
 """
 import mimetypes
-from riak.util import deprecateQuorumAccessors, deprecated
-
-
-def deprecateBucketQuorumAccessors(klass):
-    return deprecateQuorumAccessors(klass, parent='_client')
+from riak.util import lazy_property
 
 
 def bucket_property(name, doc=None):
@@ -35,7 +31,6 @@ def bucket_property(name, doc=None):
     return property(_prop_getter, _prop_setter, doc=doc)
 
 
-@deprecateBucketQuorumAccessors
 class RiakBucket(object):
     """
     The ``RiakBucket`` object allows you to access and change information
@@ -43,29 +38,38 @@ class RiakBucket(object):
     objects within the bucket.
     """
 
-    def __init__(self, client, name):
+    def __init__(self, client, name, bucket_type):
         """
         Returns a new ``RiakBucket`` instance.
 
-        :param client: A :class:`RiakClient <riak.client.RiakClient>` instance
+        :param client: A :class:`RiakClient <riak.client.RiakClient>`
+               instance
         :type client: :class:`RiakClient <riak.client.RiakClient>`
         :param name: The bucket name
         :type name: string
+        :param bucket_type: The parent bucket type of this bucket
+        :type bucket_type: :class:`BucketType`
         """
         try:
             if isinstance(name, basestring):
                 name = name.encode('ascii')
+            else:
+                raise TypeError('Bucket name must be a string')
         except UnicodeError:
             raise TypeError('Unicode bucket names are not supported.')
 
+        if not isinstance(bucket_type, BucketType):
+            raise TypeError('Parent bucket type must be a BucketType instance')
+
         self._client = client
         self.name = name
+        self.bucket_type = bucket_type
         self._encoders = {}
         self._decoders = {}
         self._resolver = None
 
     def __hash__(self):
-        return hash((self.name, self._client))
+        return hash((self.bucket_type.name, self.name, self._client))
 
     def __eq__(self, other):
         if isinstance(other, self.__class__):
@@ -137,19 +141,38 @@ class RiakBucket(object):
 
     def new(self, key=None, data=None, content_type='application/json',
             encoded_data=None):
-        """
-        Create a new :class:`RiakObject <riak.riak_object.RiakObject>`
-        that will be stored as JSON. A shortcut for manually
-        instantiating a :class:`RiakObject
-        <riak.riak_object.RiakObject>`.
+        """A shortcut for manually instantiating a new
+        :class:`~riak.riak_object.RiakObject` or a new
+        :class:`~riak.datatypes.Datatype`, based on the presence and value
+        of the :attr:`datatype <BucketType.datatype>` bucket property. When
+        the bucket contains a :class:`~riak.datatypes.Datatype`, all
+        arguments are ignored except ``key``, otherwise they are used to
+        initialize the :class:`~riak.riak_object.RiakObject`.
 
         :param key: Name of the key. Leaving this to be None (default)
                     will make Riak generate the key on store.
-        :type key: string
-        :param data: The data to store.
+        :type key: str
+        :param data: The data to store in a
+           :class:`~riak.riak_object.RiakObject`, see
+           :attr:`RiakObject.data <riak.riak_object.RiakObject.data>`.
         :type data: object
-        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
+        :param content_type: The media type of the data stored in the
+           :class:`~riak.riak_object.RiakObject`, see
+           :attr:`RiakObject.content_type
+           <riak.riak_object.RiakObject.content_type>`.
+        :type content_type: str
+        :param encoded_data: The encoded data to store in a
+           :class:`~riak.riak_object.RiakObject`, see
+           :attr:`RiakObject.encoded_data
+           <riak.riak_object.RiakObject.encoded_data>`.
+        :type encoded_data: str
+        :rtype: :class:`~riak.riak_object.RiakObject` or
+                :class:`~riak.datatypes.Datatype`
+
         """
+        if self.bucket_type.datatype:
+            return TYPES[self.bucket_type.datatype](bucket=self, key=key)
+
         try:
             if isinstance(data, basestring):
                 data = data.encode('ascii')
@@ -164,33 +187,12 @@ class RiakBucket(object):
             obj.encoded_data = encoded_data
         return obj
 
-    def new_binary(self, key=None, data=None,
-                   content_type='application/octet-stream'):
+    def get(self, key, r=None, pr=None, timeout=None, include_context=None,
+            basic_quorum=None, notfound_ok=None):
         """
-        Create a new :class:`RiakObject <riak.riak_object.RiakObject>`
-        that will be stored as plain text/binary. A shortcut for
-        manually instantiating a :class:`RiakObject
-        <riak.riak_object.RiakObject>`.
-
-        .. deprecated:: 2.0.0
-           Use :meth:`new` instead.
-
-        :param key: Name of the key.
-        :type key: string
-        :param data: The data to store.
-        :type data: object
-        :param content_type: The content type of the object.
-        :type content_type: string
-        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
-        """
-        deprecated('RiakBucket.new_binary is deprecated, '
-                   'use RiakBucket.new with the encoded_data '
-                   'param instead of data')
-        return self.new(key, encoded_data=data, content_type=content_type)
-
-    def get(self, key, r=None, pr=None, timeout=None):
-        """
-        Retrieve an object from Riak.
+        Retrieve an :class:`~riak.riak_object.RiakObject` or
+        :class:`~riak.datatypes.Datatype`, based on the presence and value
+        of the :attr:`datatype <BucketType.datatype>` bucket property.
 
         :param key: Name of the key.
         :type key: string
@@ -200,33 +202,32 @@ class RiakBucket(object):
         :type pr: integer
         :param timeout: a timeout value in milliseconds
         :type timeout: int
-        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
+        :param include_context: if the bucket contains datatypes, include
+           the opaque context in the result
+        :type include_context: bool
+        :param basic_quorum: whether to use the "basic quorum" policy
+           for not-founds
+        :type basic_quorum: bool
+        :param notfound_ok: whether to treat not-found responses as successful
+        :type notfound_ok: bool
+        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>` or
+           :class:`~riak.datatypes.Datatype`
+
         """
-        obj = RiakObject(self._client, self, key)
-        return obj.reload(r=r, pr=pr, timeout=timeout)
+        if self.bucket_type.datatype:
+            return self._client.fetch_datatype(self, key, r=r, pr=pr,
+                                               timeout=timeout,
+                                               include_context=include_context,
+                                               basic_quorum=basic_quorum,
+                                               notfound_ok=notfound_ok)
+        else:
+            obj = RiakObject(self._client, self, key)
+            return obj.reload(r=r, pr=pr, timeout=timeout,
+                              basic_quorum=basic_quorum,
+                              notfound_ok=notfound_ok)
 
-    def get_binary(self, key, r=None, pr=None, timeout=None):
-        """
-        Retrieve a binary/string object from Riak.
-
-        .. deprecated:: 2.0.0
-           Use :meth:`get` instead.
-
-        :param key: Name of the key.
-        :type key: string
-        :param r: R-Value of the request (defaults to bucket's R)
-        :type r: integer
-        :param pr: PR-Value of the request (defaults to bucket's PR)
-        :type pr: integer
-        :param timeout: a timeout value in milliseconds
-        :type timeout: int
-        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
-        """
-        deprecated('RiakBucket.get_binary is deprecated, '
-                   'use RiakBucket.get')
-        return self.get(key, r=r, pr=pr, timeout=timeout)
-
-    def multiget(self, keys, r=None, pr=None):
+    def multiget(self, keys, r=None, pr=None, timeout=None,
+                 basic_quorum=None, notfound_ok=None):
         """
         Retrieves a list of keys belonging to this bucket in parallel.
 
@@ -236,10 +237,21 @@ class RiakBucket(object):
         :type r: integer
         :param pr: PR-Value for the requests (defaults to bucket's PR)
         :type pr: integer
-        :rtype: list of :class:`RiakObject <riak.riak_object.RiakObject>`
+        :param timeout: a timeout value in milliseconds
+        :type timeout: int
+        :param basic_quorum: whether to use the "basic quorum" policy
+           for not-founds
+        :type basic_quorum: bool
+        :param notfound_ok: whether to treat not-found responses as successful
+        :type notfound_ok: bool
+        :rtype: list of :class:`RiakObjects <riak.riak_object.RiakObject>`,
+            :class:`Datatypes <riak.datatypes.Datatype>`, or tuples of
+            bucket_type, bucket, key, and the exception raised on fetch
         """
-        bkeys = [(self.name, key) for key in keys]
-        return self._client.multiget(bkeys, r=r, pr=pr)
+        bkeys = [(self.bucket_type.name, self.name, key) for key in keys]
+        return self._client.multiget(bkeys, r=r, pr=pr, timeout=timeout,
+                                     basic_quorum=basic_quorum,
+                                     notfound_ok=notfound_ok)
 
     def _get_resolver(self):
         if callable(self._resolver):
@@ -255,8 +267,8 @@ class RiakBucket(object):
         else:
             raise TypeError("resolver is not a function")
 
-    resolver = property(_get_resolver, _set_resolver, doc=
-                        """The sibling-resolution function for this
+    resolver = property(_get_resolver, _set_resolver,
+                        doc="""The sibling-resolution function for this
                            bucket. If the resolver is not set, the
                            client's resolver will be used.""")
 
@@ -369,16 +381,22 @@ class RiakBucket(object):
         """
         Streams all keys within the bucket through an iterator.
 
+        The caller must close the stream when finished.  See
+        :meth:`RiakClient.stream_keys()
+        <riak.client.RiakClient.stream_keys>` for more details.
+
         :rtype: iterator
         """
         return self._client.stream_keys(self)
 
     def new_from_file(self, key, filename):
-        """
-        Create a new Riak object in the bucket, using the contents of
+        """Create a new Riak object in the bucket, using the contents of
         the specified file. This is a shortcut for :meth:`new`, where the
         ``encoded_data`` and ``content_type`` are set for you.
 
+        .. warning:: This is not supported for buckets that contain
+           :class:`Datatypes <riak.datatypes.Datatype>`.
+
         :param key: the key of the new object
         :type key: string
         :param filename: the file to read the contents from
@@ -395,35 +413,22 @@ class RiakBucket(object):
             mimetype = 'application/octet-stream'
         return self.new(key, encoded_data=binary_data, content_type=mimetype)
 
-    def new_binary_from_file(self, key, filename):
-        """
-        Create a new Riak object in the bucket, using the contents of
-        the specified file. This is a shortcut for :meth:`new`, where the
-        ``encoded_data`` and ``content_type`` are set for you.
-
-        .. deprecated:: 2.0.0
-           Use :meth:`new_from_file` instead.
-
-        :param key: the key of the new object
-        :type key: string
-        :param filename: the file to read the contents from
-        :type filename: string
-        :rtype: :class:`RiakObject <riak.riak_object.RiakObject>`
-        """
-        deprecated('RiakBucket.new_binary_from_file is deprecated, use '
-                   'RiakBucket.new_from_file')
-        return self.new_from_file(key, filename)
-
     def search_enabled(self):
         """
         Returns True if search indexing is enabled for this
         bucket.
+
+        .. deprecated:: 2.1.0 (Riak 2.0)
+           Use :ref:`Riak Search 2.0 <yz-label>` instead.
         """
         return self.get_properties().get('search', False)
 
     def enable_search(self):
         """
         Enable search indexing for this bucket.
+
+        .. deprecated:: 2.1.0 (Riak 2.0)
+           Use :ref:`Riak Search 2.0 <yz-label>` instead.
         """
         if not self.search_enabled():
             self.set_property('search', True)
@@ -432,18 +437,29 @@ class RiakBucket(object):
     def disable_search(self):
         """
         Disable search indexing for this bucket.
+
+        .. deprecated:: 2.1.0 (Riak 2.0)
+           Use :ref:`Riak Search 2.0 <yz-label>` instead.
         """
         if self.search_enabled():
             self.set_property('search', False)
         return True
 
-    def search(self, query, **params):
+    def search(self, query, index=None, **params):
         """
         Queries a search index over objects in this bucket/index. See
         :meth:`RiakClient.fulltext_search()
         <riak.client.RiakClient.fulltext_search>` for more details.
+
+        :param query: the search query
+        :type query: string
+        :param index: the index to search over. Defaults to the bucket's name.
+        :type index: string or None
+        :param params: additional query flags
+        :type params: dict
         """
-        return self._client.fulltext_search(self.name, query, **params)
+        search_index = index or self.name
+        return self._client.fulltext_search(search_index, query, **params)
 
     def get_index(self, index, startkey, endkey=None, return_terms=None,
                   max_results=None, continuation=None, timeout=None,
@@ -454,31 +470,66 @@ class RiakBucket(object):
         :meth:`RiakClient.get_index()
         <riak.client.RiakClient.get_index>` for more details.
         """
-        return self._client.get_index(self.name, index, startkey, endkey,
+        return self._client.get_index(self, index, startkey, endkey,
                                       return_terms=return_terms,
                                       max_results=max_results,
                                       continuation=continuation,
                                       timeout=timeout, term_regex=term_regex)
 
+    def paginate_index(self, index, startkey, endkey=None,
+                       return_terms=None, max_results=1000,
+                       continuation=None, timeout=None, term_regex=None):
+        """
+        Paginates through a secondary index over objects in this bucket,
+        returning keys or index/key pairs. See
+        :meth:`RiakClient.paginate_index()
+        <riak.client.RiakClient.paginate_index>` for more details.
+        """
+        return self._client.paginate_index(self, index, startkey, endkey,
+                                           return_terms=return_terms,
+                                           max_results=max_results,
+                                           continuation=continuation,
+                                           timeout=timeout,
+                                           term_regex=term_regex)
+
     def stream_index(self, index, startkey, endkey=None, return_terms=None,
                      max_results=None, continuation=None, timeout=None,
                      term_regex=None):
         """
         Queries a secondary index over objects in this bucket,
-        streaming keys or index/key pairs via an iterator. See
+        streaming keys or index/key pairs via an iterator.
+        The caller must close the stream when finished.  See
         :meth:`RiakClient.stream_index()
         <riak.client.RiakClient.stream_index>` for more details.
         """
-        return self._client.stream_index(self.name, index, startkey, endkey,
+        return self._client.stream_index(self, index, startkey, endkey,
                                          return_terms=return_terms,
                                          max_results=max_results,
                                          continuation=continuation,
                                          timeout=timeout,
                                          term_regex=term_regex)
 
+    def paginate_stream_index(self, index, startkey, endkey=None,
+                              return_terms=None, max_results=1000,
+                              continuation=None, timeout=None,
+                              term_regex=None):
+        """
+        Paginates through a secondary index over objects in this bucket,
+        streaming keys or index/key pairs. The caller must close the stream
+        when finished.  See :meth:`RiakClient.paginate_stream_index()
+        <riak.client.RiakClient.paginate_stream_index>` for more details.
+        """
+        return self._client.paginate_stream_index(self, index, startkey,
+                                                  endkey,
+                                                  return_terms=return_terms,
+                                                  max_results=max_results,
+                                                  continuation=continuation,
+                                                  timeout=timeout,
+                                                  term_regex=term_regex)
+
     def delete(self, key, **kwargs):
-        """Deletes an object from riak. Short hand for
-        bucket.new(key).delete(). See :meth:`RiakClient.delete()
+        """Deletes a key from Riak. Short hand for
+        ``bucket.new(key).delete()``. See :meth:`RiakClient.delete()
         <riak.client.RiakClient.delete>` for options.
 
         :param key: The key for the object
@@ -493,6 +544,10 @@ class RiakBucket(object):
         :meth:`RiakClient.get_counter()
         <riak.client.RiakClient.get_counter>` for options.
 
+        .. deprecated:: 2.1.0 (Riak 2.0) Riak 1.4-style counters are
+           deprecated in favor of the :class:`~riak.datatypes.Counter`
+           datatype.
+
         :param key: the key of the counter
         :type key: string
         :rtype: int
@@ -506,6 +561,9 @@ class RiakBucket(object):
         :meth:`RiakClient.update_counter()
         <riak.client.RiakClient.update_counter>` for options.
 
+        .. deprecated:: 2.1.0 (Riak 2.0) Riak 1.4-style counters are
+           deprecated in favor of the :class:`~riak.datatypes.Counter`
+           datatype.
 
         :param key: the key of the counter
         :type key: string
@@ -517,6 +575,160 @@ class RiakBucket(object):
     increment_counter = update_counter
 
     def __str__(self):
-        return '<RiakBucket "{0}">'.format(self.name)
+        if self.bucket_type.is_default():
+            return '<RiakBucket {0!r}>'.format(self.name)
+        else:
+            return '<RiakBucket {0!r}/{1!r}>'.format(self.bucket_type.name,
+                                                     self.name)
+
+    __repr__ = __str__
+
+
+class BucketType(object):
+    """
+    The ``BucketType`` object allows you to access and change
+    properties on a Riak bucket type and access buckets within its
+    namespace.
+    """
+    def __init__(self, client, name):
+        """
+        Returns a new ``BucketType`` instance.
+
+        :param client: A :class:`RiakClient <riak.client.RiakClient>`
+               instance
+        :type client: :class:`RiakClient <riak.client.RiakClient>`
+        :param name: The bucket-type's name
+        :type name: string
+        """
+        self._client = client
+        self.name = name
+
+    def is_default(self):
+        """
+        Whether this bucket type is the default type, or a user-defined type.
+
+        :rtype: bool
+
+        """
+        return self.name == 'default'
+
+    def get_property(self, key):
+        """
+        Retrieve a bucket-type property.
+
+        :param key: The property to retrieve.
+        :type key: string
+        :rtype: mixed
+        """
+        return self.get_properties()[key]
+
+    def set_property(self, key, value):
+        """
+        Set a bucket-type property.
+
+        :param key: Property to set.
+        :type key: string
+        :param value: Property value.
+        :type value: mixed
+        """
+        return self.set_properties({key: value})
+
+    def get_properties(self):
+        """
+        Retrieve a dict of all bucket-type properties.
+
+        :rtype: dict
+        """
+        return self._client.get_bucket_type_props(self)
+
+    def set_properties(self, props):
+        """
+        Set multiple bucket-type properties in one call.
+
+        :param props: A dictionary of properties
+        :type props: dict
+        """
+        self._client.set_bucket_type_props(self, props)
+
+    def bucket(self, name):
+        """
+        Gets a bucket that belongs to this bucket-type.
+
+        :param name: the bucket name
+        :type name: str
+        :rtype: :class:`RiakBucket`
+        """
+        return self._client.bucket(name, self)
+
+    def get_buckets(self, timeout=None):
+        """
+        Get the list of buckets under this bucket-type as
+        :class:`RiakBucket <riak.bucket.RiakBucket>` instances.
+
+        .. warning:: Do not use this in production, as it requires
+           traversing through all keys stored in a cluster.
+
+        .. note:: This request is automatically retried :attr:`retries`
+           times if it fails due to network error.
+
+        :param timeout: a timeout value in milliseconds
+        :type timeout: int
+        :rtype: list of :class:`RiakBucket <riak.bucket.RiakBucket>`
+                instances
+        """
+        return self._client.get_buckets(bucket_type=self, timeout=timeout)
+
+    def stream_buckets(self, timeout=None):
+        """
+        Streams the list of buckets under this bucket-type. This is a
+        generator method that should be iterated over.
+
+        The caller must close the stream when finished.  See
+        :meth:`RiakClient.stream_buckets()
+        <riak.client.RiakClient.stream_buckets>` for more details.
+
+        .. warning:: Do not use this in production, as it requires
+           traversing through all keys stored in a cluster.
+
+        :param timeout: a timeout value in milliseconds
+        :type timeout: int
+        :rtype: iterator that yields lists of :class:`RiakBucket
+             <riak.bucket.RiakBucket>` instances
+        """
+        return self._client.stream_buckets(bucket_type=self, timeout=timeout)
+
+    @lazy_property
+    def datatype(self):
+        """
+        The assigned datatype for this bucket type, if present.
+
+        :rtype: None or string
+        """
+        if self.is_default():
+            return None
+        else:
+            return self.get_properties().get('datatype')
+
+    def __str__(self):
+        return "<BucketType {0!r}>".format(self.name)
+
+    __repr__ = __str__
+
+    def __hash__(self):
+        return hash((self.name, self._client))
+
+    def __eq__(self, other):
+        if isinstance(other, self.__class__):
+            return hash(self) == hash(other)
+        else:
+            return False
+
+    def __ne__(self, other):
+        if isinstance(other, self.__class__):
+            return hash(self) != hash(other)
+        else:
+            return True
+
 
 from riak_object import RiakObject
+from riak.datatypes import TYPES
