@@ -1,22 +1,19 @@
 # -*- coding: utf-8 -*-
-from __future__ import with_statement
-
-import os
 import random
 import platform
+from threading import Thread
+from Queue import Queue
 
 if platform.python_version() < '2.7':
     unittest = __import__('unittest2')
 else:
     import unittest
 
+from riak import RiakError
 from riak.client import RiakClient
-from riak.mapreduce import RiakKeyFilter
-from riak import key_filter
 from riak.riak_object import RiakObject
 
-from riak.test_server import TestServer
-
+from riak.tests.test_yokozuna import YZSearchTests
 from riak.tests.test_search import SearchTests, \
     EnableSearchTests, SolrSearchTests
 from riak.tests.test_mapreduce import MapReduceAliasTests, \
@@ -24,61 +21,92 @@ from riak.tests.test_mapreduce import MapReduceAliasTests, \
 from riak.tests.test_kv import BasicKVTests, KVFileTests, \
     BucketPropsTest, CounterTests
 from riak.tests.test_2i import TwoITests
+from riak.tests.test_btypes import BucketTypeTests
+from riak.tests.test_security import SecurityTests
+from riak.tests.test_datatypes import DatatypeIntegrationTests
 
-try:
-    __import__('riak_pb')
-    HAVE_PROTO = True
-except ImportError:
-    HAVE_PROTO = False
-
-HOST = os.environ.get('RIAK_TEST_HOST', '127.0.0.1')
-
-PB_HOST = os.environ.get('RIAK_TEST_PB_HOST', HOST)
-PB_PORT = int(os.environ.get('RIAK_TEST_PB_PORT', '8087'))
-
-HTTP_HOST = os.environ.get('RIAK_TEST_HTTP_HOST', HOST)
-HTTP_PORT = int(os.environ.get('RIAK_TEST_HTTP_PORT', '8098'))
-
-USE_TEST_SERVER = int(os.environ.get('USE_TEST_SERVER', '0'))
-
-if USE_TEST_SERVER:
-    HTTP_PORT = 9000
-    PB_PORT = 9002
-    test_server = TestServer()
-    test_server.cleanup()
-    test_server.prepare()
-    test_server.start()
+from riak.tests import HOST, PB_HOST, PB_PORT, HTTP_HOST, HTTP_PORT, \
+    HAVE_PROTO, DUMMY_HTTP_PORT, DUMMY_PB_PORT, \
+    SKIP_SEARCH, RUN_YZ, SECURITY_CREDS, SKIP_POOL
 
 testrun_search_bucket = None
 testrun_props_bucket = None
 testrun_sibs_bucket = None
+testrun_yz = {'btype': None, 'bucket': None, 'index': None}
+testrun_yz_index = {'btype': None, 'bucket': None, 'index': None}
+testrun_yz_mr = {'btype': None, 'bucket': None, 'index': None}
 
 
 def setUpModule():
     global testrun_search_bucket, testrun_props_bucket, \
-        testrun_sibs_bucket
+        testrun_sibs_bucket, testrun_yz, testrun_yz_index, testrun_yz_mr
 
-    c = RiakClient(protocol='http', host=HTTP_HOST, http_port=HTTP_PORT)
+    c = RiakClient(host=PB_HOST, http_port=HTTP_PORT,
+                   pb_port=PB_PORT, credentials=SECURITY_CREDS)
 
     testrun_props_bucket = 'propsbucket'
     testrun_sibs_bucket = 'sibsbucket'
     c.bucket(testrun_sibs_bucket).allow_mult = True
 
-    if not int(os.environ.get('SKIP_SEARCH', '0')):
+    if (not SKIP_SEARCH and not RUN_YZ):
         testrun_search_bucket = 'searchbucket'
         b = c.bucket(testrun_search_bucket)
         b.enable_search()
 
+    if RUN_YZ:
+        # YZ index on bucket of the same name
+        testrun_yz = {'btype': None, 'bucket': 'yzbucket',
+                      'index': 'yzbucket'}
+        # YZ index on bucket of a different name
+        testrun_yz_index = {'btype': None, 'bucket': 'yzindexbucket',
+                            'index': 'yzindex'}
+        # Add bucket and type for Search 2.0 -> MapReduce
+        testrun_yz_mr = {'btype': 'pytest-mr', 'bucket': 'mrbucket',
+                         'index': 'mrbucket'}
+
+        for yz in (testrun_yz, testrun_yz_index, testrun_yz_mr):
+            c.create_search_index(yz['index'])
+            if yz['btype'] is not None:
+                t = c.bucket_type(yz['btype'])
+                b = t.bucket(yz['bucket'])
+            else:
+                b = c.bucket(yz['bucket'])
+            # Keep trying to set search bucket property until it succeeds
+            index_set = False
+            while not index_set:
+                try:
+                    b.set_property('search_index', yz['index'])
+                    index_set = True
+                except RiakError:
+                    pass
+
 
 def tearDownModule():
-    c = RiakClient(protocol='http', host=HTTP_HOST, http_port=HTTP_PORT)
-    if not int(os.environ.get('SKIP_SEARCH', '0')):
+    global testrun_search_bucket, testrun_props_bucket, \
+        testrun_sibs_bucket, testrun_yz_bucket
+
+    c = RiakClient(host=HTTP_HOST, http_port=HTTP_PORT,
+                   pb_port=PB_PORT, credentials=SECURITY_CREDS)
+
+    c.bucket(testrun_sibs_bucket).clear_properties()
+    c.bucket(testrun_props_bucket).clear_properties()
+
+    if not SKIP_SEARCH and not RUN_YZ:
         b = c.bucket(testrun_search_bucket)
         b.clear_properties()
-    b = c.bucket(testrun_sibs_bucket)
-    b.clear_properties()
-    b = c.bucket(testrun_props_bucket)
-    b.clear_properties()
+
+    if RUN_YZ:
+        for yz in (testrun_yz, testrun_yz_index, testrun_yz_mr):
+            if yz['btype'] is not None:
+                t = c.bucket_type(yz['btype'])
+                b = t.bucket(yz['bucket'])
+            else:
+                b = c.bucket(yz['bucket'])
+            b.set_property('search_index', '_dont_index_')
+            c.delete_search_index(yz['index'])
+            for keys in b.stream_keys():
+                for key in keys:
+                    b.delete(key)
 
 
 class BaseTestCase(object):
@@ -86,6 +114,7 @@ class BaseTestCase(object):
     host = None
     pb_port = None
     http_port = None
+    credentials = None
 
     @staticmethod
     def randint():
@@ -99,14 +128,17 @@ class BaseTestCase(object):
         return out
 
     def create_client(self, host=None, http_port=None, pb_port=None,
-                      protocol=None, **client_args):
+                      protocol=None, credentials=None,
+                      **client_args):
         host = host or self.host or HOST
         http_port = http_port or self.http_port or HTTP_PORT
         pb_port = pb_port or self.pb_port or PB_PORT
         protocol = protocol or self.protocol
+        credentials = credentials or SECURITY_CREDS
         return RiakClient(protocol=protocol,
                           host=host,
                           http_port=http_port,
+                          credentials=credentials,
                           pb_port=pb_port, **client_args)
 
     def setUp(self):
@@ -115,6 +147,10 @@ class BaseTestCase(object):
         self.search_bucket = testrun_search_bucket
         self.sibs_bucket = testrun_sibs_bucket
         self.props_bucket = testrun_props_bucket
+        self.yz = testrun_yz
+        self.yz_index = testrun_yz_index
+        self.yz_mr = testrun_yz_mr
+        self.credentials = SECURITY_CREDS
 
         self.client = self.create_client()
 
@@ -123,12 +159,41 @@ class ClientTests(object):
     def test_request_retries(self):
         # We guess at some ports that will be unused by Riak or
         # anything else.
-        client = self.create_client(http_port=1023, pb_port=1022)
+        client = self.create_client(http_port=DUMMY_HTTP_PORT,
+                                    pb_port=DUMMY_PB_PORT)
 
         # If retries are exhausted, the final result should also be an
         # error.
         self.assertRaises(IOError, client.ping)
 
+    def test_request_retries_configurable(self):
+        # We guess at some ports that will be unused by Riak or
+        # anything else.
+        client = self.create_client(http_port=DUMMY_HTTP_PORT,
+                                    pb_port=DUMMY_PB_PORT)
+
+        # Change the retry count
+        client.retries = 10
+        self.assertEqual(10, client.retries)
+
+        # The retry count should be a thread local
+        retries = Queue()
+
+        def _target():
+            retries.put(client.retries)
+            retries.join()
+
+        th = Thread(target=_target)
+        th.start()
+        self.assertEqual(3, retries.get(block=True))
+        retries.task_done()
+        th.join()
+
+        # Modify the retries in a with statement
+        with client.retry_count(5):
+            self.assertEqual(5, client.retries)
+            self.assertRaises(IOError, client.ping)
+
     def test_timeout_validation(self):
         bucket = self.client.bucket(self.bucket_name)
         key = self.key_name
@@ -194,25 +259,63 @@ class ClientTests(object):
         and not propagated.
         """
         keys = [self.key_name, self.randname(), self.randname()]
-        client = self.create_client(http_port=1023, pb_port=1024)
+        client = self.create_client(http_port=DUMMY_HTTP_PORT,
+                                    pb_port=DUMMY_PB_PORT)
         results = client.bucket(self.bucket_name).multiget(keys)
         for failure in results:
             self.assertIsInstance(failure, tuple)
-            self.assertEqual(failure[0], self.bucket_name)
-            self.assertIn(failure[1], keys)
-            self.assertIsInstance(failure[2], StandardError)
+            self.assertEqual(failure[0], 'default')
+            self.assertEqual(failure[1], self.bucket_name)
+            self.assertIn(failure[2], keys)
+            self.assertIsInstance(failure[3], StandardError)
 
     def test_multiget_notfounds(self):
         """
         Not founds work in multiget just the same as get.
         """
-        keys = [(self.bucket_name, self.key_name),
-                (self.bucket_name, self.randname())]
+        keys = [("default", self.bucket_name, self.key_name),
+                ("default", self.bucket_name, self.randname())]
         results = self.client.multiget(keys)
         for obj in results:
             self.assertIsInstance(obj, RiakObject)
             self.assertFalse(obj.exists)
 
+    def test_multiget_pool_size(self):
+        """
+        The pool size for multigets can be configured at client initiation
+        time. Multiget still works as expected.
+        """
+        client = self.create_client(multiget_pool_size=2)
+        self.assertEqual(2, client._multiget_pool._size)
+
+        keys = [self.key_name, self.randname(), self.randname()]
+        for key in keys:
+            client.bucket(self.bucket_name)\
+                .new(key, encoded_data=key, content_type="text/plain")\
+                .store()
+
+        results = client.bucket(self.bucket_name).multiget(keys)
+        for obj in results:
+            self.assertIsInstance(obj, RiakObject)
+            self.assertTrue(obj.exists)
+            self.assertEqual(obj.key, obj.encoded_data)
+
+    @unittest.skipIf(SKIP_POOL, 'SKIP_POOL is set')
+    def test_pool_close(self):
+        """
+        Iterate over the connection pool and close all connections.
+        """
+        # Do something to add to the connection pool
+        self.test_multiget_bucket()
+        if self.client.protocol == 'pbc':
+            self.assertGreater(len(self.client._pb_pool.resources), 1)
+        else:
+            self.assertGreater(len(self.client._http_pool.resources), 1)
+        # Now close them all up
+        self.client.close()
+        self.assertEqual(len(self.client._http_pool.resources), 0)
+        self.assertEqual(len(self.client._pb_pool.resources), 0)
+
 
 class RiakPbcTransportTestCase(BasicKVTests,
                                KVFileTests,
@@ -225,8 +328,12 @@ class RiakPbcTransportTestCase(BasicKVTests,
                                MapReduceStreamTests,
                                EnableSearchTests,
                                SearchTests,
+                               YZSearchTests,
                                ClientTests,
                                CounterTests,
+                               BucketTypeTests,
+                               SecurityTests,
+                               DatatypeIntegrationTests,
                                BaseTestCase,
                                unittest.TestCase):
 
@@ -236,8 +343,6 @@ class RiakPbcTransportTestCase(BasicKVTests,
         self.host = PB_HOST
         self.pb_port = PB_PORT
         self.protocol = 'pbc'
-        self.http_client = self.create_client(HTTP_HOST,
-                                              http_port=HTTP_PORT)
         super(RiakPbcTransportTestCase, self).setUp()
 
     def test_uses_client_id_if_given(self):
@@ -258,8 +363,12 @@ class RiakHttpTransportTestCase(BasicKVTests,
                                 EnableSearchTests,
                                 SolrSearchTests,
                                 SearchTests,
+                                YZSearchTests,
                                 ClientTests,
                                 CounterTests,
+                                BucketTypeTests,
+                                SecurityTests,
+                                DatatypeIntegrationTests,
                                 BaseTestCase,
                                 unittest.TestCase):
 
@@ -286,63 +395,5 @@ class RiakHttpTransportTestCase(BasicKVTests,
         self.assertEqual(len(stored_object.links), 400)
 
 
-class FilterTests(unittest.TestCase):
-    def test_simple(self):
-        f1 = RiakKeyFilter("tokenize", "-", 1)
-        self.assertEqual(f1._filters, [["tokenize", "-", 1]])
-
-    def test_add(self):
-        f1 = RiakKeyFilter("tokenize", "-", 1)
-        f2 = RiakKeyFilter("eq", "2005")
-        f3 = f1 + f2
-        self.assertEqual(list(f3), [["tokenize", "-", 1], ["eq", "2005"]])
-
-    def test_and(self):
-        f1 = RiakKeyFilter("starts_with", "2005-")
-        f2 = RiakKeyFilter("ends_with", "-01")
-        f3 = f1 & f2
-        self.assertEqual(list(f3),
-                         [["and",
-                           [["starts_with", "2005-"]],
-                           [["ends_with", "-01"]]]])
-
-    def test_multi_and(self):
-        f1 = RiakKeyFilter("starts_with", "2005-")
-        f2 = RiakKeyFilter("ends_with", "-01")
-        f3 = RiakKeyFilter("matches", "-11-")
-        f4 = f1 & f2 & f3
-        self.assertEqual(list(f4), [["and",
-                                     [["starts_with", "2005-"]],
-                                     [["ends_with", "-01"]],
-                                     [["matches", "-11-"]],
-                                     ]])
-
-    def test_or(self):
-        f1 = RiakKeyFilter("starts_with", "2005-")
-        f2 = RiakKeyFilter("ends_with", "-01")
-        f3 = f1 | f2
-        self.assertEqual(list(f3), [["or", [["starts_with", "2005-"]],
-                                     [["ends_with", "-01"]]]])
-
-    def test_multi_or(self):
-        f1 = RiakKeyFilter("starts_with", "2005-")
-        f2 = RiakKeyFilter("ends_with", "-01")
-        f3 = RiakKeyFilter("matches", "-11-")
-        f4 = f1 | f2 | f3
-        self.assertEqual(list(f4), [["or",
-                                     [["starts_with", "2005-"]],
-                                     [["ends_with", "-01"]],
-                                     [["matches", "-11-"]],
-                                     ]])
-
-    def test_chaining(self):
-        f1 = key_filter.tokenize("-", 1).eq("2005")
-        f2 = key_filter.tokenize("-", 2).eq("05")
-        f3 = f1 & f2
-        self.assertEqual(list(f3), [["and",
-                                     [["tokenize", "-", 1], ["eq", "2005"]],
-                                     [["tokenize", "-", 2], ["eq", "05"]]
-                                     ]])
-
 if __name__ == '__main__':
     unittest.main()
