@@ -19,6 +19,7 @@ import riak_pb
 from riak import RiakError
 from riak.content import RiakContent
 from riak.util import decode_index_value
+from riak.multidict import MultiDict
 
 
 def _invert(d):
@@ -49,11 +50,30 @@ QUORUM_TO_PY = _invert(QUORUM_TO_PB)
 
 NORMAL_PROPS = ['n_val', 'allow_mult', 'last_write_wins', 'old_vclock',
                 'young_vclock', 'big_vclock', 'small_vclock', 'basic_quorum',
-                'notfound_ok', 'search', 'backend']
+                'notfound_ok', 'search', 'backend', 'search_index', 'datatype']
 COMMIT_HOOK_PROPS = ['precommit', 'postcommit']
 MODFUN_PROPS = ['chash_keyfun', 'linkfun']
 QUORUM_PROPS = ['r', 'pr', 'w', 'pw', 'dw', 'rw']
 
+MAP_FIELD_TYPES = {
+    riak_pb.MapField.COUNTER: 'counter',
+    riak_pb.MapField.SET: 'set',
+    riak_pb.MapField.REGISTER: 'register',
+    riak_pb.MapField.FLAG: 'flag',
+    riak_pb.MapField.MAP: 'map',
+    'counter': riak_pb.MapField.COUNTER,
+    'set': riak_pb.MapField.SET,
+    'register': riak_pb.MapField.REGISTER,
+    'flag': riak_pb.MapField.FLAG,
+    'map': riak_pb.MapField.MAP
+}
+
+DT_FETCH_TYPES = {
+    riak_pb.DtFetchResp.COUNTER: 'counter',
+    riak_pb.DtFetchResp.SET: 'set',
+    riak_pb.DtFetchResp.MAP: 'map'
+}
+
 
 class RiakPbcCodec(object):
     """
@@ -399,7 +419,8 @@ class RiakPbcCodec(object):
         :type timeout: int
         :rtype riak_pb.RpbIndexReq
         """
-        req = riak_pb.RpbIndexReq(bucket=bucket, index=index)
+        req = riak_pb.RpbIndexReq(bucket=bucket.name, index=index)
+        self._add_bucket_type(req, bucket.bucket_type)
         if endkey:
             req.qtype = riak_pb.RpbIndexReq.range
             req.range_min = str(startkey)
@@ -421,3 +442,162 @@ class RiakPbcCodec(object):
         if term_regex:
             req.term_regex = term_regex
         return req
+
+    def _decode_search_index(self, index):
+        """
+        Fills an RpbYokozunaIndex message with the appropriate data.
+
+        :param index: a yz index message
+        :type index: riak_pb.RpbYokozunaIndex
+        :rtype dict
+        """
+        result = {}
+        result['name'] = index.name
+        if index.HasField('schema'):
+            result['schema'] = index.schema
+        if index.HasField('n_val'):
+            result['n_val'] = index.n_val
+        return result
+
+    def _add_bucket_type(self, req, bucket_type):
+        if bucket_type and not bucket_type.is_default():
+            if not self.bucket_types():
+                raise NotImplementedError(
+                    'Server does not support bucket-types')
+            req.type = bucket_type.name
+
+    def _encode_search_query(self, req, params):
+        if 'rows' in params:
+            req.rows = params['rows']
+        if 'start' in params:
+            req.start = params['start']
+        if 'sort' in params:
+            req.sort = params['sort']
+        if 'filter' in params:
+            req.filter = params['filter']
+        if 'df' in params:
+            req.df = params['df']
+        if 'op' in params:
+            req.op = params['op']
+        if 'q.op' in params:
+            req.op = params['q.op']
+        if 'fl' in params:
+            if isinstance(params['fl'], list):
+                req.fl.extend(params['fl'])
+            else:
+                req.fl.append(params['fl'])
+        if 'presort' in params:
+            req.presort = params['presort']
+
+    def _decode_search_doc(self, doc):
+        resultdoc = MultiDict()
+        for pair in doc.fields:
+            ukey = unicode(pair.key, 'utf-8')
+            uval = unicode(pair.value, 'utf-8')
+            resultdoc.add(ukey, uval)
+        return resultdoc.mixed()
+
+    def _decode_dt_fetch(self, resp):
+        dtype = DT_FETCH_TYPES.get(resp.type)
+        if dtype is None:
+            raise ValueError("Unknown datatype on wire: {}".format(resp.type))
+
+        value = self._decode_dt_value(dtype, resp.value)
+
+        if resp.HasField('context'):
+            context = resp.context[:]
+        else:
+            context = None
+
+        return dtype, value, context
+
+    def _decode_dt_value(self, dtype, msg):
+        if dtype == 'counter':
+            return msg.counter_value
+        elif dtype == 'set':
+            return self._decode_set_value(msg.set_value)
+        elif dtype == 'map':
+            return self._decode_map_value(msg.map_value)
+
+    def _encode_dt_options(self, req, params):
+        for q in ['r', 'pr', 'w', 'dw', 'pw']:
+            if q in params and params[q] is not None:
+                setattr(req, q, self._encode_quorum(params[q]))
+
+        for o in ['basic_quorum', 'notfound_ok', 'timeout', 'return_body',
+                  'include_context']:
+            if o in params and params[o] is not None:
+                setattr(req, o, params[o])
+
+    def _decode_map_value(self, entries):
+        out = {}
+        for entry in entries:
+            name = entry.field.name[:]
+            dtype = MAP_FIELD_TYPES[entry.field.type]
+            if dtype == 'counter':
+                value = entry.counter_value
+            elif dtype == 'set':
+                value = self._decode_set_value(entry.set_value)
+            elif dtype == 'register':
+                value = entry.register_value[:]
+            elif dtype == 'flag':
+                value = entry.flag_value
+            elif dtype == 'map':
+                value = self._decode_map_value(entry.map_value)
+            out[(name, dtype)] = value
+        return out
+
+    def _decode_set_value(self, set_value):
+        return [string[:] for string in set_value]
+
+    def _encode_dt_op(self, dtype, req, op):
+        if dtype == 'counter':
+            req.op.counter_op.increment = op[1]
+        elif dtype == 'set':
+            self._encode_set_op(req.op, op)
+        elif dtype == 'map':
+            self._encode_map_op(req.op.map_op, op)
+        else:
+            raise TypeError("Cannot send operation on datatype {!r}".
+                            format(dtype))
+
+    def _encode_set_op(self, msg, op):
+        if 'adds' in op:
+            msg.set_op.adds.extend(op['adds'])
+        if 'removes' in op:
+            msg.set_op.removes.extend(op['removes'])
+
+    def _encode_map_op(self, msg, ops):
+        for op in ops:
+            name, dtype = op[1]
+            ftype = MAP_FIELD_TYPES[dtype]
+            if op[0] == 'add':
+                add = msg.adds.add()
+                add.name = name
+                add.type = ftype
+            elif op[0] == 'remove':
+                remove = msg.removes.add()
+                remove.name = name
+                remove.type = ftype
+            elif op[0] == 'update':
+                update = msg.updates.add()
+                update.field.name = name
+                update.field.type = ftype
+                self._encode_map_update(dtype, update, op[2])
+
+    def _encode_map_update(self, dtype, msg, op):
+        if dtype == 'counter':
+            # ('increment', some_int)
+            msg.counter_op.increment = op[1]
+        elif dtype == 'set':
+            self._encode_set_op(msg, op)
+        elif dtype == 'map':
+            self._encode_map_op(msg.map_op, op)
+        elif dtype == 'register':
+            # ('assign', some_str)
+            msg.register_op = op[1]
+        elif dtype == 'flag':
+            if op == 'enable':
+                msg.flag_op = riak_pb.MapUpdate.ENABLE
+            else:
+                msg.flag_op = riak_pb.MapUpdate.DISABLE
