@@ -29,7 +29,7 @@ from stream import (RiakPbcKeyStream, RiakPbcMapredStream, RiakPbcBucketStream,
                     RiakPbcIndexStream)
 from codec import RiakPbcCodec
 
-from messages import (
+from riak_pb.messages import (
     MSG_CODE_PING_REQ,
     MSG_CODE_PING_RESP,
     MSG_CODE_GET_CLIENT_ID_REQ,
@@ -51,7 +51,9 @@ from messages import (
     MSG_CODE_GET_BUCKET_RESP,
     MSG_CODE_SET_BUCKET_REQ,
     MSG_CODE_SET_BUCKET_RESP,
-    MSG_CODE_MAPRED_REQ,
+    MSG_CODE_GET_BUCKET_TYPE_REQ,
+    MSG_CODE_SET_BUCKET_TYPE_REQ,
+    MSG_CODE_MAP_RED_REQ,
     MSG_CODE_INDEX_REQ,
     MSG_CODE_INDEX_RESP,
     MSG_CODE_SEARCH_QUERY_REQ,
@@ -61,7 +63,18 @@ from messages import (
     MSG_CODE_COUNTER_UPDATE_REQ,
     MSG_CODE_COUNTER_UPDATE_RESP,
     MSG_CODE_COUNTER_GET_REQ,
-    MSG_CODE_COUNTER_GET_RESP
+    MSG_CODE_COUNTER_GET_RESP,
+    MSG_CODE_YOKOZUNA_INDEX_GET_REQ,
+    MSG_CODE_YOKOZUNA_INDEX_GET_RESP,
+    MSG_CODE_YOKOZUNA_INDEX_PUT_REQ,
+    MSG_CODE_YOKOZUNA_INDEX_DELETE_REQ,
+    MSG_CODE_YOKOZUNA_SCHEMA_GET_REQ,
+    MSG_CODE_YOKOZUNA_SCHEMA_GET_RESP,
+    MSG_CODE_YOKOZUNA_SCHEMA_PUT_REQ,
+    MSG_CODE_DT_FETCH_REQ,
+    MSG_CODE_DT_FETCH_RESP,
+    MSG_CODE_DT_UPDATE_REQ,
+    MSG_CODE_DT_UPDATE_RESP
 )
 
 
@@ -71,7 +84,11 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
     buffers interface on the riak server.
     """
 
-    def __init__(self, node=None, client=None, timeout=None, *unused_options):
+    def __init__(self,
+                 node=None,
+                 client=None,
+                 timeout=None,
+                 *unused_options):
         """
         Construct a new RiakPbcTransport object.
         """
@@ -123,7 +140,8 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
     client_id = property(_get_client_id, _set_client_id,
                          doc="""the client ID for this connection""")
 
-    def get(self, robj, r=None, pr=None, timeout=None):
+    def get(self, robj, r=None, pr=None, timeout=None, basic_quorum=None,
+            notfound_ok=None):
         """
         Serialize get request and deserialize response
         """
@@ -132,21 +150,26 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
         req = riak_pb.RpbGetReq()
         if r:
             req.r = self._encode_quorum(r)
-        if self.quorum_controls() and pr:
-            req.pr = self._encode_quorum(pr)
+        if self.quorum_controls():
+            if pr:
+                req.pr = self._encode_quorum(pr)
+            if basic_quorum is not None:
+                req.basic_quorum = basic_quorum
+            if notfound_ok is not None:
+                req.notfound_ok = notfound_ok
         if self.client_timeouts() and timeout:
             req.timeout = timeout
         if self.tombstone_vclocks():
-            req.deletedvclock = 1
+            req.deletedvclock = True
 
         req.bucket = bucket.name
+        self._add_bucket_type(req, bucket.bucket_type)
+
         req.key = robj.key
 
         msg_code, resp = self._request(MSG_CODE_GET_REQ, req,
                                        MSG_CODE_GET_RESP)
 
-        # TODO: support if_modified flag
-
         if resp is not None:
             if resp.HasField('vclock'):
                 robj.vclock = VClock(resp.vclock, 'binary')
@@ -162,9 +185,6 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
 
     def put(self, robj, w=None, dw=None, pw=None, return_body=True,
             if_none_match=False, timeout=None):
-        """
-        Serialize get request and deserialize response
-        """
         bucket = robj.bucket
 
         req = riak_pb.RpbPutReq()
@@ -183,6 +203,8 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
             req.timeout = timeout
 
         req.bucket = bucket.name
+        self._add_bucket_type(req, bucket.bucket_type)
+
         if robj.key:
             req.key = robj.key
         if robj.vclock:
@@ -207,11 +229,6 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
 
     def delete(self, robj, rw=None, r=None, w=None, dw=None, pr=None, pw=None,
                timeout=None):
-        """
-        Serialize get request and deserialize response
-        """
-        bucket = robj.bucket
-
         req = riak_pb.RpbDelReq()
         if rw:
             req.rw = self._encode_quorum(rw)
@@ -234,7 +251,9 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
         if self.tombstone_vclocks() and robj.vclock:
             req.vclock = robj.vclock.encode('binary')
 
+        bucket = robj.bucket
         req.bucket = bucket.name
+        self._add_bucket_type(req, bucket.bucket_type)
         req.key = robj.key
 
         msg_code, resp = self._request(MSG_CODE_DEL_REQ, req,
@@ -259,6 +278,7 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
         """
         req = riak_pb.RpbListKeysReq()
         req.bucket = bucket.name
+        self._add_bucket_type(req, bucket.bucket_type)
         if self.client_timeouts() and timeout:
             req.timeout = timeout
 
@@ -266,20 +286,21 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
 
         return RiakPbcKeyStream(self)
 
-    def get_buckets(self, timeout=None):
+    def get_buckets(self, bucket_type=None, timeout=None):
         """
         Serialize bucket listing request and deserialize response
         """
-        req = None
+        req = riak_pb.RpbListBucketsReq()
+        self._add_bucket_type(req, bucket_type)
+
         if self.client_timeouts() and timeout:
-            req = riak_pb.RpbListBucketsReq()
             req.timeout = timeout
 
         msg_code, resp = self._request(MSG_CODE_LIST_BUCKETS_REQ, req,
                                        MSG_CODE_LIST_BUCKETS_RESP)
         return resp.buckets
 
-    def stream_buckets(self, timeout=None):
+    def stream_buckets(self, bucket_type=None, timeout=None):
         """
         Stream list of buckets through an iterator
         """
@@ -290,6 +311,7 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
 
         req = riak_pb.RpbListBucketsReq()
         req.stream = True
+        self._add_bucket_type(req, bucket_type)
         # Bucket streaming landed in the same release as timeouts, so
         # we don't need to check the capability.
         if timeout:
@@ -305,6 +327,7 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
         """
         req = riak_pb.RpbGetBucketReq()
         req.bucket = bucket.name
+        self._add_bucket_type(req, bucket.bucket_type)
 
         msg_code, resp = self._request(MSG_CODE_GET_BUCKET_REQ, req,
                                        MSG_CODE_GET_BUCKET_RESP)
@@ -317,6 +340,7 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
         """
         req = riak_pb.RpbSetBucketReq()
         req.bucket = bucket.name
+        self._add_bucket_type(req, bucket.bucket_type)
 
         if not self.pb_all_bucket_props():
             for key in props:
@@ -339,10 +363,40 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
 
         req = riak_pb.RpbResetBucketReq()
         req.bucket = bucket.name
+        self._add_bucket_type(req, bucket.bucket_type)
         self._request(MSG_CODE_RESET_BUCKET_REQ, req,
                       MSG_CODE_RESET_BUCKET_RESP)
         return True
 
+    def get_bucket_type_props(self, bucket_type):
+        """
+        Fetch bucket-type properties
+        """
+        self._check_bucket_types(bucket_type)
+
+        req = riak_pb.RpbGetBucketTypeReq()
+        req.type = bucket_type.name
+
+        msg_code, resp = self._request(MSG_CODE_GET_BUCKET_TYPE_REQ, req,
+                                       MSG_CODE_GET_BUCKET_RESP)
+
+        return self._decode_bucket_props(resp.props)
+
+    def set_bucket_type_props(self, bucket_type, props):
+        """
+        Set bucket-type properties
+        """
+        self._check_bucket_types(bucket_type)
+
+        req = riak_pb.RpbSetBucketTypeReq()
+        req.type = bucket_type.name
+
+        self._encode_bucket_props(props, req)
+
+        msg_code, resp = self._request(MSG_CODE_SET_BUCKET_TYPE_REQ, req,
+                                       MSG_CODE_SET_BUCKET_RESP)
+        return True
+
     def mapred(self, inputs, query, timeout=None):
         # dictionary of phase results - each content should be an encoded array
         # which is appended to the result for that phase.
@@ -370,7 +424,7 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
         req.request = content
         req.content_type = "application/json"
 
-        self._send_msg(MSG_CODE_MAPRED_REQ, req)
+        self._send_msg(MSG_CODE_MAP_RED_REQ, req)
 
         return RiakPbcMapredStream(self)
 
@@ -397,7 +451,7 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
         else:
             results = resp.keys[:]
 
-        if max_results:
+        if max_results is not None and resp.HasField('continuation'):
             return (results, resp.continuation)
         else:
             return (results, None)
@@ -422,32 +476,89 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
 
         return RiakPbcIndexStream(self, index, return_terms)
 
+    def create_search_index(self, index, schema=None, n_val=None):
+        if not self.pb_search_admin():
+            raise NotImplementedError("Search 2.0 administration is not "
+                                      "supported for this version")
+        idx = riak_pb.RpbYokozunaIndex(name=index)
+        if schema:
+            idx.schema = schema
+        if n_val:
+            idx.n_val = n_val
+        req = riak_pb.RpbYokozunaIndexPutReq(index=idx)
+
+        self._request(MSG_CODE_YOKOZUNA_INDEX_PUT_REQ, req,
+                      MSG_CODE_PUT_RESP)
+        return True
+
+    def get_search_index(self, index):
+        if not self.pb_search_admin():
+            raise NotImplementedError("Search 2.0 administration is not "
+                                      "supported for this version")
+        req = riak_pb.RpbYokozunaIndexGetReq(name=index)
+
+        msg_code, resp = self._request(MSG_CODE_YOKOZUNA_INDEX_GET_REQ, req,
+                                       MSG_CODE_YOKOZUNA_INDEX_GET_RESP)
+        if len(resp.index) > 0:
+            return self._decode_search_index(resp.index[0])
+        else:
+            raise RiakError('notfound')
+
+    def list_search_indexes(self):
+        if not self.pb_search_admin():
+            raise NotImplementedError("Search 2.0 administration is not "
+                                      "supported for this version")
+        req = riak_pb.RpbYokozunaIndexGetReq()
+
+        msg_code, resp = self._request(MSG_CODE_YOKOZUNA_INDEX_GET_REQ, req,
+                                       MSG_CODE_YOKOZUNA_INDEX_GET_RESP)
+
+        return [self._decode_search_index(index) for index in resp.index]
+
+    def delete_search_index(self, index):
+        if not self.pb_search_admin():
+            raise NotImplementedError("Search 2.0 administration is not "
+                                      "supported for this version")
+        req = riak_pb.RpbYokozunaIndexDeleteReq(name=index)
+
+        self._request(MSG_CODE_YOKOZUNA_INDEX_DELETE_REQ, req,
+                      MSG_CODE_DEL_RESP)
+
+        return True
+
+    def create_search_schema(self, schema, content):
+        if not self.pb_search_admin():
+            raise NotImplementedError("Search 2.0 administration is not "
+                                      "supported for this version")
+        scma = riak_pb.RpbYokozunaSchema(name=schema, content=content)
+        req = riak_pb.RpbYokozunaSchemaPutReq(schema=scma)
+
+        self._request(MSG_CODE_YOKOZUNA_SCHEMA_PUT_REQ, req,
+                      MSG_CODE_PUT_RESP)
+        return True
+
+    def get_search_schema(self, schema):
+        if not self.pb_search_admin():
+            raise NotImplementedError("Search 2.0 administration is not "
+                                      "supported for this version")
+        req = riak_pb.RpbYokozunaSchemaGetReq(name=schema)
+
+        msg_code, resp = self._request(MSG_CODE_YOKOZUNA_SCHEMA_GET_REQ, req,
+                                       MSG_CODE_YOKOZUNA_SCHEMA_GET_RESP)
+        result = {}
+        result['name'] = resp.schema.name
+        result['content'] = resp.schema.content
+        return result
+
     def search(self, index, query, **params):
         if not self.pb_search():
             return self._search_mapred_emu(index, query)
 
+        if isinstance(query, unicode):
+            query = query.encode('utf8')
+
         req = riak_pb.RpbSearchQueryReq(index=index, q=query)
-        if 'rows' in params:
-            req.rows = params['rows']
-        if 'start' in params:
-            req.start = params['start']
-        if 'sort' in params:
-            req.sort = params['sort']
-        if 'filter' in params:
-            req.filter = params['filter']
-        if 'df' in params:
-            req.df = params['df']
-        if 'op' in params:
-            req.op = params['op']
-        if 'q.op' in params:
-            req.op = params['q.op']
-        if 'fl' in params:
-            if isinstance(params['fl'], list):
-                req.fl.extend(params['fl'])
-            else:
-                req.fl.append(params['fl'])
-        if 'presort' in params:
-            req.presort = params['presort']
+        self._encode_search_query(req, params)
 
         msg_code, resp = self._request(MSG_CODE_SEARCH_QUERY_REQ, req,
                                        MSG_CODE_SEARCH_QUERY_RESP)
@@ -457,18 +568,15 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
             result['max_score'] = resp.max_score
         if resp.HasField('num_found'):
             result['num_found'] = resp.num_found
-        docs = []
-        for doc in resp.docs:
-            resultdoc = {}
-            for pair in doc.fields:
-                ukey = unicode(pair.key, 'utf-8')
-                uval = unicode(pair.value, 'utf-8')
-                resultdoc[ukey] = uval
-            docs.append(resultdoc)
-        result['docs'] = docs
+        result['docs'] = [self._decode_search_doc(doc) for doc in resp.docs]
         return result
 
     def get_counter(self, bucket, key, **params):
+        if not bucket.bucket_type.is_default():
+            raise NotImplementedError("Counters are not "
+                                      "supported with bucket-types, "
+                                      "use datatypes instead.")
+
         if not self.counters():
             raise NotImplementedError("Counters are not supported")
 
@@ -492,6 +600,11 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
             return None
 
     def update_counter(self, bucket, key, value, **params):
+        if not bucket.bucket_type.is_default():
+            raise NotImplementedError("Counters are not "
+                                      "supported with bucket-types, "
+                                      "use datatypes instead.")
+
         if not self.counters():
             raise NotImplementedError("Counters are not supported")
 
@@ -514,3 +627,63 @@ class RiakPbcTransport(RiakTransport, RiakPbcConnection, RiakPbcCodec):
             return resp.value
         else:
             return True
+
+    def fetch_datatype(self, bucket, key, **options):
+
+        if bucket.bucket_type.is_default():
+            raise NotImplementedError("Datatypes cannot be used in the default"
+                                      " bucket-type.")
+
+        if not self.datatypes():
+            raise NotImplementedError("Datatypes are not supported.")
+
+        req = riak_pb.DtFetchReq()
+        req.type = bucket.bucket_type.name
+        req.bucket = bucket.name
+        req.key = key
+        self._encode_dt_options(req, options)
+
+        msg_code, resp = self._request(MSG_CODE_DT_FETCH_REQ, req,
+                                       MSG_CODE_DT_FETCH_RESP)
+
+        return self._decode_dt_fetch(resp)
+
+    def update_datatype(self, datatype, **options):
+
+        if datatype.bucket.bucket_type.is_default():
+            raise NotImplementedError("Datatypes cannot be used in the default"
+                                      " bucket-type.")
+
+        if not self.datatypes():
+            raise NotImplementedError("Datatypes are not supported.")
+
+        op = datatype.to_op()
+        type_name = datatype.type_name
+        if not op:
+            raise ValueError("No operation to send on datatype {!r}".
+                             format(datatype))
+
+        req = riak_pb.DtUpdateReq()
+        req.bucket = datatype.bucket.name
+        req.type = datatype.bucket.bucket_type.name
+
+        if datatype.key:
+            req.key = datatype.key
+        if datatype._context:
+            req.context = datatype._context
+
+        self._encode_dt_options(req, options)
+
+        self._encode_dt_op(type_name, req, op)
+
+        msg_code, resp = self._request(MSG_CODE_DT_UPDATE_REQ, req,
+                                       MSG_CODE_DT_UPDATE_RESP)
+        if resp.HasField('key'):
+            datatype.key = resp.key[:]
+        if resp.HasField('context'):
+            datatype._context = resp.context[:]
+
+        if options.get('return_body'):
+            datatype._set_value(self._decode_dt_value(type_name, resp))
+
+        return True
